"""
Functions for broadening spectra
"""
from typing import Callable, Literal, Union
import warnings

import numpy as np
from numpy.polynomial import Chebyshev
from pint import Quantity
from scipy.optimize import nnls
from scipy.stats import norm
from scipy.signal import convolve

from euphonic import ureg


ErrorFit = Literal['cheby-log', 'cubic']
KernelShape = Literal['gauss', 'lorentz']


def variable_width_broadening(
    bins: Quantity,
    x: Quantity,
    width_function: Callable[[Quantity], Quantity],
    weights: Union[np.ndarray, Quantity],
    width_lower_limit: Quantity = None,
    width_convention: Literal['fwhm', 'std'] = 'fwhm',
    adaptive_error: float = 1e-2,
    shape: KernelShape = 'gauss',
    fit: ErrorFit = 'cheby-log'
    ) -> Quantity:
    r"""Apply x-dependent Gaussian broadening to 1-D data series

    Typically this is an energy-dependent instrumental resolution function.
    Data is binned and broadened to output array with reciprocal units.

    A fast interpolation-based method is used to reduce the number of Gaussian
    evaluations.

    Parameters
    ----------
    bins
        Data bins for output spectrum
    x
        Data positions (to be binned)
    width_function
        A function handle which takes an n-dimensional array of x values as
        input and returns corresponding width values for broadening. These
        should be Quantity arrays and dimensionally-consistent with x.
    weights
        Weight for each data point corresponding to x. Note that these should
        be "counts" rather than binned spectral weights; this function will
        bin the data and apply bin-width weighting.
    width_lower_limit
        A lower bound is set for broadening width in WIDTH_UNIT. If set to None
        (default) the bin width will be used. To disable any lower limit, set
        to 0 or lower.
    width_convention
        Indicate if polynomial function yields standard deviation (sigma) or
        full-width half-maximum.
    adaptive_error
        Acceptable error for gaussian approximations, defined
        as the absolute difference between the areas of the true and
        approximate gaussians.
    shape
        Select broadening kernel function.
    fit
        Select parametrisation of kernel width spacing to adaptive_error.
        'cheby-log' is recommended: for shape 'gauss', 'cubic' is also
        available.
    """

    if width_convention.lower() == 'fwhm' and shape == 'gauss':
        sigma_function = (lambda x: width_function(x)
                          / np.sqrt(8 * np.log(2)))
    elif width_convention.lower() == 'std' and shape == 'lorentz':
        raise ValueError('Standard deviation unavailable for Lorentzian '
                         'function: please use FWHM.')
    elif width_convention.lower() in ('std', 'fwhm'):
        sigma_function = width_function
    else:
        raise ValueError('width_convention must be "std" or "fwhm".')

    widths = sigma_function(x)

    if width_lower_limit is None:
        width_lower_limit = np.diff(bins).max()

    widths = np.maximum(widths, width_lower_limit)

    if isinstance(weights, np.ndarray):
        weights = weights * ureg('dimensionless')
        assert isinstance(weights, Quantity)

    weights_unit = weights.units

    return width_interpolated_broadening(bins, x, widths,
                                         weights.magnitude,
                                         adaptive_error=adaptive_error,
                                         shape=shape,
                                         fit=fit) * weights_unit


def width_interpolated_broadening(
    bins: Quantity,
    x: Quantity,
    widths: Quantity,
    weights: np.ndarray,
    adaptive_error: float,
    shape: KernelShape = 'gauss',
    fit: ErrorFit = 'cheby-log'
    ) -> Quantity:
    """
    Uses a fast, approximate method to broaden a spectrum
    with a variable-width kernel. Exact Gaussians are calculated
    at logrithmically spaced values across the range of widths.
    A small set of spectra that have been scaled using the weights
    from linear combinations of the exact Gaussians are convolved
    using Fast Fourier Transforms (FFT) and then summed to give the
    approximate broadened spectra.

    Parameters
    ----------
    bins
        The energy bin edges to use for calculating
        the spectrum
    x
        Broadening samples
    widths
        The broadening width for each peak, must be the same shape as x.
    weights
        The weight for each peak, must be the same shape as x.
    adaptive_error
        Scalar float. Acceptable error for gaussian approximations, defined
        as the absolute difference between the areas of the true and
        approximate gaussians.
    shape
        Select kernel shape. Widths will correspond to sigma or gamma
        parameters respectively.
    fit
        Select parametrisation of kernel width spacing to adaptive_error.
        'cheby-log' is recommended: for shape 'gauss', 'cubic' is also
        available.

    Returns
    -------
    spectrum
        Quantity of shape (bins - 1,) containing broadened spectrum
        ydata
    """

    conv = 1*ureg('hartree').to(bins.units)
    return _width_interpolated_broadening(bins.to('hartree').magnitude,
                                          x.to('hartree').magnitude,
                                          widths.to('hartree').magnitude,
                                          weights,
                                          adaptive_error,
                                          shape=shape,
                                          fit=fit) / conv


def _lorentzian(x: np.ndarray, gamma: np.ndarray) -> np.ndarray:
    return gamma / (2 * np.pi * (x**2 + (gamma / 2)**2))


def _get_spacing(error,
                 shape: KernelShape = 'gauss',
                 fit: ErrorFit = 'cheby-log'):
    """
    Determine suitable spacing value for mode_width given accepted error level

    Coefficients have been fitted to plots of error vs spacing value
    """

    if fit == 'cubic' and shape == 'gauss':
        return np.polyval([612.7, -122.7, 15.40, 1.0831], error)

    elif fit == 'cheby-log':
        if shape == 'lorentz':
            cheby = Chebyshev(
                [1.26039672, 0.39900457, 0.20392176, 0.08602507,
                 0.03337662, 0.00878684, 0.00619626],
                window=[1., 1.],
                domain=[-4.99146317, -1.34655197])
            safe_domain = [-4, -1.35]

        else:  # gauss
            cheby = Chebyshev(
                [1.25885858, 0.39803148, 0.20311735, 0.08654827,
                 0.03447873, 0.00894006, 0.00715706],
                window=[-1., 1.],
                domain=[-4.64180022, -1.00029948])
            safe_domain = [-4, -1.]

        log_error = np.log10(error)
        if log_error < safe_domain[0] or log_error > safe_domain[1]:
            raise ValueError("Target error is out of fit range; value must lie"
                             f" in range {np.power(10, safe_domain)}.")
        return cheby(log_error)

    else:
        raise ValueError(f'Fit "{fit}" is not available for shape "{shape}". '
                         f'The "cheby-log" fit is recommended for "gauss" and '
                         f'"Lorentz" shapes.')


def _width_interpolated_broadening(
    bins: np.ndarray,
    x: np.ndarray,
    widths: np.ndarray,
    weights: np.ndarray,
    adaptive_error: float,
    shape: KernelShape = 'gauss',
    fit: ErrorFit = 'cheby-log') -> np.ndarray:
    """
    Broadens a spectrum using a variable-width kernel, taking the
    same arguments as `variable_width` but expects arrays with
    consistent units rather than Quantities. Also returns an array
    rather than a Quantity.
    """
    x = np.ravel(x)
    widths = np.ravel(widths)
    weights = np.ravel(weights)
    spacing = _get_spacing(adaptive_error, shape=shape, fit=fit)

    # bins should be regularly spaced, check that this is the case and
    # raise a warning if not
    bin_widths = np.diff(bins)
    if not np.all(np.isclose(bin_widths, bin_widths[0])):
        warnings.warn('Not all bin widths are equal, so broadening by '
                      'convolution will give incorrect results.',
                      stacklevel=3)
    bin_width = bin_widths[0]

    n_kernels = int(
        np.ceil(np.log(max(widths)/min(widths))/np.log(spacing)))
    width_samples = spacing**np.arange(n_kernels+1)*min(widths)

    # Evaluate kernels on regular grid of length equal to number of bins,
    #  avoids the need for zero padding in convolution step
    if (len(bins) % 2) == 0:
        x_values = np.arange(-len(bins)/2+1, len(bins)/2)*bin_width
    else:
        x_values = np.arange(-int(len(bins)/2), int(len(bins)/2)+1)*bin_width

    if shape == 'gauss':
        kernels = norm.pdf(x_values, scale=width_samples[:, np.newaxis]
                           ) * bin_width
    elif shape == 'lorentz':
        kernels = _lorentzian(x_values, gamma=width_samples[:, np.newaxis]
                              ) * bin_width

    kernels_idx = np.searchsorted(width_samples, widths, side="right")

    if len(kernels) == 1:
        # Only one kernel to consider, use 100% in all cases
        lower_coeffs = [1.]
    else:
        lower_coeffs = find_coeffs(spacing, shape=shape)

    spectrum = np.zeros(len(bins)-1)

    for i in range(1, len(width_samples)+1):
        masked_block = (kernels_idx == i)
        width_factors = widths[masked_block]/width_samples[i-1]
        lower_mix = np.polyval(lower_coeffs, width_factors)
        lower_weights = lower_mix * weights[masked_block]

        if i == 1:
            hist, _ = np.histogram(x[masked_block], bins=bins,
                                   weights=lower_weights/bin_width)
        else:
            mixing_weights = np.concatenate((upper_weights_prev,
                                             lower_weights))
            hist_x = np.concatenate((x_prev, x[masked_block]))
            hist, _ = np.histogram(hist_x, bins=bins,
                                   weights=mixing_weights/bin_width)

        x_prev = x[masked_block]
        upper_weights_prev = weights[masked_block] - lower_weights

        spectrum += convolve(hist, kernels[i-1], mode="same", method="fft")

    return spectrum


def find_coeffs(spacing: float,
                shape: KernelShape = 'gauss') -> np.ndarray:
    """"
    Function that, for a given spacing value, gives the coefficients of the
    polynomial which describes the relationship between kernel width and the
    linear combination weights determined by optimised interpolation

    Parameters
    ----------
    spacing
        Scalar float. The spacing value between sigma (or gamma) samples at
        which the kernel is exactly calculated.
    shape
        'gauss' or 'lorentz', selecting the type of broadening kernel.

    Returns
    -------
    coeffs
        Array containing the polynomial coefficients, with the highest
        power first
    """
    width_values = np.linspace(1, spacing, num=10)
    x_range = np.linspace(-10, 10, num=101)
    if shape == 'gauss':
        actual_kernels = norm.pdf(x_range, scale=width_values[:, np.newaxis])
    elif shape == 'lorentz':
        actual_kernels = _lorentzian(x_range,
                                     gamma=width_values[:, np.newaxis])
    lower_mix = np.zeros(len(width_values))
    ref_kernels = actual_kernels[[0, -1]].T

    # For each width value, use non-negative least squares fitting to
    # find the linear combination weights that best reproduce the
    # actual kernel.
    for i in range(len(width_values)):
        actual_kernel = actual_kernels[i]
        res = nnls(ref_kernels, actual_kernel)[0]
        lower_mix[i] = res[0]

    coeffs = np.polyfit(width_values, lower_mix, 3)
    return coeffs



from multiprocessing import cpu_count
from typing import Union, Optional, Dict, Any, Type, TypeVar

import spglib as spg
import numpy as np
try:
    import brille as br
except ModuleNotFoundError as err:
    raise ModuleNotFoundError(
        'Cannot import Brille for use with BrilleInterpolator '
        '(maybe Brille is not installed?). To install Euphonic\'s '
        'optional Brille dependency, try:\n\npip install '
        'euphonic[brille]\n') from err

from euphonic.validate import _check_constructor_inputs
from euphonic import (ureg, QpointPhononModes, QpointFrequencies,
                      ForceConstants, Crystal)

class BrilleInterpolator:
    """
    A class to perform linear interpolation of eigenvectors and
    frequencies at arbitrary q-points using the Brille library

    Attributes
    ----------
    crystal : Crystal
        Lattice and atom information
    """
    T = TypeVar('T', bound='BrilleInterpolator')

    def __init__(self, crystal: Crystal,
                 grid: Union[br.BZTrellisQdc, br.BZMeshQdc,
                             br.BZNestQdc]) -> None:
        """
        Parameters
        ----------
        crystal
            Lattice and atom information
        grid
            A Brille grid filled with phonon eigenvectors in
            Cartesian coordinates and phonon frequencies in Hartree. See
            Brille documentation for details.
        """
        _check_constructor_inputs(
            [crystal, grid],
            [Crystal, [br.BZTrellisQdc, br.BZMeshQdc, br.BZNestQdc]],
            [(), ()],
            ['crystal', 'grid'])
        # Check grid has been filled and vals/vecs are the correct shape
        n_atoms = crystal.n_atoms
        n_qpts = len(grid.rlu)
        _check_constructor_inputs(
            [grid.values, grid.vectors],
            [np.ndarray, np.ndarray],
            [(n_qpts, 3*n_atoms, 1), (n_qpts, 3*n_atoms, 3*n_atoms)],
            ['grid.values', 'grid.vectors'])

        self._grid = grid
        self.crystal = crystal

    def calculate_qpoint_phonon_modes(self, qpts: np.ndarray, **kwargs
                                      ) -> QpointPhononModes:
        """
        Calculate phonon frequencies and eigenvectors at specified
        q-points via linear interpolation

        Parameters
        ----------
        qpts
            Shape (n_qpts, 3) float ndarray. The q-points to
            interpolate onto in reciprocal cell vector units.
        **kwargs
            Will be passed to the
            brille.BZTrellis/Mesh/Nest.ir_interpolate_at
            method. By default useparallel=True and
            threads=multiprocessing.cpu_count() are passed.

        Returns
        -------
        qpoint_phonon_modes
            An object containing frequencies and eigenvectors
            linearly interpolated at each q-point
        """
        vals, vecs = self._br_grid_calculate_phonons(qpts, **kwargs)
        return QpointPhononModes(
            self.crystal, qpts, vals, vecs)

    def calculate_qpoint_frequencies(self, qpts: np.ndarray, **kwargs
                                      ) -> QpointFrequencies:
        """
        Calculate phonon frequencies at specified
        q-points via linear interpolation

        Parameters
        ----------
        qpts
            Shape (n_qpts, 3) float ndarray. The q-points to
            interpolate onto in reciprocal cell vector units.
        **kwargs
            Will be passed to the
            brille.BZTrellis/Mesh/Nest.ir_interpolate_at
            method. By default useparallel=True and
            threads=multiprocessing.cpu_count() are passed.

        Returns
        -------
        qpoint_frequencies
            An object containing frequencies linearly interpolated
            at each q-point
        """
        vals, _ = self._br_grid_calculate_phonons(qpts, **kwargs)
        return QpointFrequencies(
            self.crystal, qpts, vals)

    def _br_grid_calculate_phonons(self, qpts, **kwargs):
        """
        Pass kwargs to Brille grid ir_interpolate_at and transform
        output into Euphonic format
        """
        qpts = np.ascontiguousarray(qpts)
        if not kwargs:
            kwargs = {'useparallel': True, 'threads': cpu_count()}
        vals, vecs = self._grid.ir_interpolate_at(qpts, **kwargs)
        vals = vals.squeeze(axis=-1)*ureg('hartree').to('meV')
        n_atoms = self.crystal.n_atoms
        vecs = vecs.view().reshape(-1, 3*n_atoms, n_atoms, 3)
        return vals, vecs

    @classmethod
    def from_force_constants(
            cls: Type[T], force_constants: ForceConstants,
            grid_type: str = 'trellis', grid_npts: int = 1000,
            grid_density: Optional[int] = None,
            grid_kwargs: Optional[Dict[str, Any]] = None,
            interpolation_kwargs: Optional[Dict[str, Any]] = None) -> T:
        """
        Generates a grid over the irreducible Brillouin Zone to be
        used for linear interpolation with Brille, with properties
        determined by the grid_type, grid_npts and grid_kwargs
        arguments. Then uses ForceConstants to fill the grid points
        with phonon frequencies and eigenvectors via Fourier
        interpolation. This returns a BrilleInterpolator object that
        can then be used for linear interpolation.

        Parameters
        ----------
        grid_type
            The Brille grid type to be used, one of {'trellis',
            'mesh', 'nest'}, creating a brille.BZTrellisQdc,
            brille.BZMeshQdc or brille.BZNestQdc grid
            respectively
        grid_npts
            The approximate number of q-points in the Brille grid.
            This is used to set the kwargs for the grid creation so
            that a grid with approximately the desired number of
            points is created, note this number is approximate
            as the number of grid points generated depends on
            Brille's internal algorithm. If this number is higher,
            the linear interpolation is likely to give values
            closer to the ForceConstants Fourier interpolation,
            but the initialisation and memory costs will be higher.
            This does nothing if grid_kwargs is set.
        grid_density
            The approximate number density of q-points per
            1/angstrom^3 volume
        grid_kwargs
            Kwargs to be passed to the grid constructor (e.g.
            brille.BZTrellisQdc). If set grid_npts does
            nothing.
        interpolation_kwargs
            Kwargs to be passed to
            ForceConstants.calculate_qpoint_phonon_modes. Note
            that insert_gamma and reduce_qpoints are incompatible
            and will be ignored.

        Returns
        -------
        brille_interpolator
        """
        grid_type_opts = ('trellis', 'mesh', 'nest')
        if grid_type not in grid_type_opts:
            raise ValueError(f'Grid type "{grid_type}" not recognised')

        crystal = force_constants.crystal
        cell = crystal.to_spglib_cell()

        dataset = spg.get_symmetry_dataset(cell)
        rotations = dataset['rotations']  # in fractional
        translations = dataset['translations']  # in fractional

        symmetry = br.Symmetry(rotations, translations)
        basis = br.Basis(cell[1], cell[2])
        lattice = br.Lattice(cell[0], symmetry=symmetry, basis=basis,
                             snap_to_symmetry=True)
        # snap_to_symmetry will take care of slightly off-symmetry atom
        # basis positions and vectors
        bz = br.BrillouinZone(lattice)

        print('Generating grid...')
        vol = bz.ir_polyhedron.volume
        if grid_type == 'trellis':
            if grid_kwargs is None:
                if grid_density is not None:
                    grid_kwargs = {
                        'node_volume_fraction': 1/grid_density}
                else:
                    # node_volume_fraction actually describes cube
                    # volume used to generate tetrahedra
                    grid_kwargs = {
                        'node_volume_fraction': vol/grid_npts}
                grid_kwargs = {**grid_kwargs,
                               'always_triangulate': False}
            grid = br.BZTrellisQdc(bz, **grid_kwargs)
        elif grid_type == 'mesh':
            if grid_kwargs is None:
                if grid_density is not None:
                    grid_kwargs = {
                        'max_size': 1/grid_density,
                        'max_points': int(grid_density*vol)}
                else:
                    grid_kwargs = {
                        'max_size': vol/grid_npts,
                        'max_points': grid_npts}
            grid = br.BZMeshQdc(bz, **grid_kwargs)
        elif grid_type == 'nest':
            if grid_kwargs is None:
                if grid_density is not None:
                    grid_kwargs = {'number_density': int(grid_density*vol)}
                else:
                    grid_kwargs = {'number_density': grid_npts}
            grid = br.BZNestQdc(bz, **grid_kwargs)

        print(f'Grid generated with {len(grid.rlu)} q-points. '
               'Calculating frequencies/eigenvectors...')
        if interpolation_kwargs is None:
            interpolation_kwargs = {}
        interpolation_kwargs['insert_gamma'] = False
        interpolation_kwargs['reduce_qpts'] = False
        phonons = force_constants.calculate_qpoint_phonon_modes(
            grid.rlu, **interpolation_kwargs)

        n_atoms = crystal.n_atoms
        frequencies = np.reshape(phonons._frequencies,
                                 phonons._frequencies.shape + (1,))
        freq_el = (1,)
        freq_weight = (1., 0., 0.)
        evecs = np.reshape(phonons.eigenvectors,
                           (*phonons.eigenvectors.shape[:2], 3*n_atoms))
        n_elems = (0, 3*n_atoms, 0) # num of scalar, vector, matrix elements
        rotates_like = 2 # rotates like gamma
        length_unit = 1 # angstrom units, i.e. Cartesian eigenvectors
        evecs_el = (*n_elems, rotates_like, length_unit)
        evecs_weight = (0., 1., 0.)
        print('Filling grid...')
        grid.fill(frequencies, freq_el, freq_weight, evecs, evecs_el,
                  evecs_weight)
        return cls(force_constants.crystal, grid)



from collections import defaultdict
import re
import struct
from typing import (Any, BinaryIO, Dict, Iterator, List, NamedTuple,
                    Optional, TextIO, Tuple, Union)

import numpy as np

from euphonic import ureg


def read_phonon_dos_data(
        filename: str,
        cell_vectors_unit: str = 'angstrom',
        atom_mass_unit: str = 'amu',
        frequencies_unit: str = 'meV',
        mode_gradients_unit: str = 'meV*angstrom') -> Dict[str, Any]:
    """
    Reads data from a .phonon_dos file and returns it in a dictionary

    Parameters
    ----------
    filename
        The path and name of the .phonon file to read
    cell_vectors_unit
        The unit to return the cell vectors in
    atom_mass_unit
        The unit to return the atom masses in
    frequencies_unit
        The unit to return the frequencies in

    Returns
    -------
    data_dict : dict
        A dict with the following keys: 'n_atoms', 'cell_vectors',
        'cell_vectors_unit', 'atom_r', 'atom_type', 'atom_mass',
        'atom_mass_unit', 'qpts', 'weights', 'frequencies',
        'frequencies_unit', 'mode_gradients', 'mode_gradients_unit',
        'dos_bins', 'dos_bins_unit', 'dos', 'pdos'
    """
    with open(filename, 'r') as f:

        f.readline()  # Skip BEGIN header
        n_atoms = int(f.readline().split()[-1])
        _ = int(f.readline().split()[-1])  # n_species is not used
        n_branches = int(f.readline().split()[-1])
        n_bins = int(f.readline().split()[-1])
        (cell_vectors, atom_r, atom_type, atom_mass) = _read_crystal_info(
            f, n_atoms)
        f.readline()  # Skip BEGIN GRADIENTS

        qpts = np.empty((0, 3))
        weights = np.empty((0,))
        freqs = np.empty((0, n_branches))
        mode_grads = np.empty((0, n_branches))
        for frequency_block in _read_frequency_blocks(
                f,
                n_branches,
                extra_columns=[0],
                terminator=' END GRADIENTS\n'):

            qmode_grad = frequency_block.extra

            qpts = np.concatenate((qpts, [frequency_block.qpt]))
            weights = np.concatenate((weights, [frequency_block.weight]))
            freqs = np.concatenate((freqs, [frequency_block.frequencies]))
            mode_grads = np.concatenate((mode_grads, [qmode_grad.squeeze()]))
        line = f.readline()
        if 'BEGIN DOS' not in line:
            raise RuntimeError(
                f'Expected "BEGIN DOS" in {filename}, got {line}')

        dos_data = np.loadtxt(f, max_rows=n_bins)

    data_dict: Dict[str, Any] = {}
    cry_dict = data_dict['crystal'] = {}
    cry_dict['n_atoms'] = n_atoms
    cry_dict['cell_vectors'] = (cell_vectors*ureg('angstrom').to(
        cell_vectors_unit)).magnitude
    cry_dict['cell_vectors_unit'] = cell_vectors_unit
    cry_dict['atom_r'] = atom_r
    cry_dict['atom_type'] = atom_type
    cry_dict['atom_mass'] = atom_mass*(ureg('amu')).to(
        atom_mass_unit).magnitude
    cry_dict['atom_mass_unit'] = atom_mass_unit

    data_dict['qpts'] = qpts
    data_dict['weights'] = weights
    data_dict['frequencies'] = ((freqs*(1/ureg.cm)).to(
        frequencies_unit)).magnitude
    data_dict['frequencies_unit'] = frequencies_unit
    # Pint conversion 'angstrom*(1/cm)' to 'angstrom*(meV)' doesn't
    # work so convert separately
    mode_grads = mode_grads*ureg('angstrom').to(cell_vectors_unit).magnitude
    data_dict['mode_gradients'] = mode_grads*ureg('1/cm').to(
            frequencies_unit).magnitude
    data_dict['mode_gradients_unit'] = (frequencies_unit + '*'
                                        + cell_vectors_unit)
    data_dict['dos_bins'] = dos_data[:, 0]*ureg('1/cm').to(
            frequencies_unit).magnitude
    data_dict['dos_bins_unit'] = frequencies_unit

    data_dict['dos'] = {}
    data_dict['dos_unit'] = frequencies_unit
    # Avoid issues in converting DOS, Pint allows
    # cm^-1 -> meV but not cm -> 1/meV
    dos_conv = (1*ureg('1/cm').to(frequencies_unit)).magnitude
    dos_dict = data_dict['dos']
    dos_dict['Total'] = dos_data[:, 1]/dos_conv
    _, idx = np.unique(atom_type, return_index=True)
    unique_types = atom_type[np.sort(idx)]
    for i, species in enumerate(unique_types):
        dos_dict[species] = dos_data[:, i + 2]/dos_conv

    return data_dict


def read_phonon_data(
        filename: str,
        cell_vectors_unit: str = 'angstrom',
        atom_mass_unit: str = 'amu',
        frequencies_unit: str = 'meV',
        read_eigenvectors: bool = True,
        average_repeat_points: bool = True,
        prefer_non_loto: bool = False
    ) -> Dict[str, Any]:
    """
    Reads data from a .phonon file and returns it in a dictionary

    Parameters
    ----------
    filename
        The path and name of the .phonon file to read
    cell_vectors_unit
        The unit to return the cell vectors in
    atom_mass_unit
        The unit to return the atom masses in
    frequencies_unit
        The unit to return the frequencies in
    read_eigenvectors
        Whether to read the eigenvectors and return them in the
        dictionary
    average_repeat_points
        If multiple frequency/eigenvectors blocks are provided with the same
        q-point index (i.e. for Gamma-point with LO-TO splitting), scale the
        weights such that these sum to the given weight
    prefer_non_loto
        If multiple frequency/eigenvector blocks are provided with the same
        q-point index and all-but-one include a direction vector, use the data
        from the point without a direction vector. (i.e. use the "exact" Gamma
        data without non-analytic correction.) This option takes priority over
        average_repeat_points.

    Returns
    -------
    data_dict : dict
        A dict with the following keys: 'n_atoms', 'cell_vectors',
        'cell_vectors_unit', 'atom_r', 'atom_type', 'atom_mass',
        'atom_mass_unit', 'qpts', 'weights', 'frequencies',
        'frequencies_unit'

        If read_eigenvectors is True, there will also be an
        'eigenvectors' key
    """
    with open(filename, 'r') as f:

        f.readline()  # Skip BEGIN header
        n_atoms = int(f.readline().split()[-1])
        n_branches = int(f.readline().split()[-1])
        n_qpts = int(f.readline().split()[-1])
        (cell_vectors, atom_r, atom_type, atom_mass) = _read_crystal_info(
            f, n_atoms)

        qpts = np.zeros((n_qpts, 3))
        weights = np.zeros(n_qpts)
        freqs = np.zeros((n_qpts, n_branches))
        if read_eigenvectors:
            eigenvecs = np.zeros((n_qpts, n_branches, n_atoms, 3),
                                 dtype='complex128')

        # Need to loop through file using while rather than number of
        # q-points as points are duplicated on LO-TO splitting
        idx = 0
        previous_qpt_id = -1
        repeated_qpt_ids = defaultdict(set)
        loto_split_indices = set()

        for frequency_block in _read_frequency_blocks(f, n_branches):
            qpt_id = frequency_block.qpt_id

            if prefer_non_loto and frequency_block.direction is not None:
                loto_split_indices.add(idx)

            if ((average_repeat_points or prefer_non_loto)
                and (qpt_id == previous_qpt_id)):
                repeated_qpt_ids[qpt_id].update({idx - 1, idx})
            previous_qpt_id = qpt_id

            [f.readline() for x in range(2)]  # Skip 2 label lines
            evec_lines = np.array(
                [f.readline() for x in range(n_atoms*n_branches)])
            if read_eigenvectors:
                evec_lines = np.array([x.split()[2:] for x in evec_lines],
                                      dtype=np.float64)
                lines_i = np.column_stack((
                    [evec_lines[:, 0] + evec_lines[:, 1]*1j,
                     evec_lines[:, 2] + evec_lines[:, 3]*1j,
                     evec_lines[:, 4] + evec_lines[:, 5]*1j]))
                qeigenvec = np.zeros((n_branches, n_atoms, 3),
                                     dtype=np.complex128)
                for i in range(n_branches):
                    qeigenvec[i, :, :] = lines_i[i*n_atoms:(i+1)*n_atoms, :]
            # Sometimes there are more than n_qpts q-points in the file
            # due to LO-TO splitting
            if idx < len(qpts):
                qpts[idx] = frequency_block.qpt
                freqs[idx] = frequency_block.frequencies
                weights[idx] = frequency_block.weight
                if read_eigenvectors:
                    eigenvecs[idx] = qeigenvec
            else:
                qpts = np.concatenate((qpts, [frequency_block.qpt]))
                freqs = np.concatenate((freqs, [frequency_block.frequencies]))
                weights = np.concatenate((weights, [frequency_block.weight]))
                if read_eigenvectors:
                    eigenvecs = np.concatenate((eigenvecs, [qeigenvec]))
            idx += 1

    # Multiple qpts with same CASTEP q-pt index: correct weights
    for qpt_id, indices in repeated_qpt_ids.items():
        if (prefer_non_loto
            and (intersection := indices & loto_split_indices)
            and len(indices) > len(intersection)):
            # Repeated q-point has both split and un-split variations;
            # set weights of split points to zero
            if len(indices) - len(intersection) == 1:
                indices = np.asarray(list(intersection), dtype=int)
                weights[indices] = 0
            else:
                raise ValueError(
                    "Found multiple non-split blocks for q-point {qpt_id},"
                    " cannot determine which to use.")

        elif average_repeat_points:
            indices = np.asarray(list(indices), dtype=int)
            weights[indices] /= indices.size

    data_dict: Dict[str, Any] = {}
    data_dict['crystal'] = {}
    cry_dict = data_dict['crystal']
    cry_dict['n_atoms'] = n_atoms
    cry_dict['cell_vectors'] = (cell_vectors*ureg('angstrom').to(
        cell_vectors_unit)).magnitude
    cry_dict['cell_vectors_unit'] = cell_vectors_unit
    cry_dict['atom_r'] = atom_r
    cry_dict['atom_type'] = atom_type
    cry_dict['atom_mass'] = atom_mass*(ureg('amu')).to(
        atom_mass_unit).magnitude
    cry_dict['atom_mass_unit'] = atom_mass_unit
    data_dict['qpts'] = qpts
    data_dict['weights'] = weights
    data_dict['frequencies'] = ((freqs*(1/ureg.cm)).to(
        frequencies_unit)).magnitude
    data_dict['frequencies_unit'] = frequencies_unit
    if read_eigenvectors:
        data_dict['eigenvectors'] = eigenvecs

    return data_dict


def _read_crystal_info(
        f: TextIO, n_atoms: int
        ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Reads the header crystal information from a CASTEP text file, from
    'Unit cell vectors' to 'END header'

    Parameters
    ----------
    f
        File object in read mode for the .phonon file containing the
        data
    n_atoms
        Number of atoms in the cell

    Returns
    -------
    cell_vectors
        Shape (3, 3) float ndarray. The unit cell vectors in Angstroms
    atom_r
        Shape (n_atoms, 3) float ndarray. The fractional position of
        each atom within the unit cell
    atom_type
        Shape (n_atoms,) string ndarray. The chemical symbols of each
        atom in the unit cell
    atom_mass
        Shape (n_atoms,) float ndarray. The mass of each atom in the
        unit cell in atomic mass units
    """
    while 'Unit cell vectors' not in f.readline():
        pass
    cell_vectors = np.array([[float(x) for x in f.readline().split()[0:3]]
                             for i in range(3)])
    f.readline()  # Skip fractional co-ordinates label
    atom_info = np.array([f.readline().split() for i in range(n_atoms)])
    atom_r = np.array([[float(x) for x in y[1:4]] for y in atom_info])
    atom_type = np.array([x[4] for x in atom_info])
    atom_mass = np.array([float(x[5]) for x in atom_info])
    f.readline()  # Skip END header line

    return cell_vectors, atom_r, atom_type, atom_mass


_qpt_index_pattern = re.compile(r'q-pt=\s*(\d+)')
_qpt_float_pattern = re.compile(r'-?\d+\.\d+')


class _FrequencyBlock(NamedTuple):
    qpt_id: int
    qpt: np.ndarray
    direction: Optional[np.ndarray]
    weight: float
    frequencies: np.ndarray
    extra: Optional[np.ndarray]


def _read_frequency_block(
    f: TextIO,
    n_branches: int,
    extra_columns: Optional[List] = None,
    terminator: str = ''
        ) -> Union[_FrequencyBlock, None]:
    """
    For a single q-point reads the q-point, weight, frequencies
    and optionally any extra columns

    Parameters
    ----------
    f
        File object in read mode for the file containing the data
    n_branches
        Expected number of frequencies (i.e. phonon branches)
    extra_columns
        The index(es) of extra columns to read after the frequencies
        column. e.g. to read the first column after frequencies
        extra_columns = [0]. To read the first and second columns after
        frequencies extra_columns = [0, 1]. For reference, in .phonon
        IR intensities = 0, Raman intensities = 1, and in .phonon_dos
        mode gradients = 0
    terminator
        If this line is reached, return None without raising an error

    Returns
    -------
    NamedTuple with attributes:

    qpt_id
        CASTEP-assigned index of the q-point
    qpt
        Shape (3,) float ndarray. The q-point in reciprocal fractional
        coordinates
    direction
        If q-point is a Gamma-point with non-analytic LO-TO splitting
        correction, this is the direction vector of that data. Otherwise, None.
    weight
        The weight of this q-point
    frequencies
        Shape (n_modes,) float ndarray. The phonon frequencies for this
        q-point in 1/cm
    extra
        None if extra_columns is None, otherwise a shape
        (len(extra_columns), n_modes) float ndarray
    """
    qpt_line = f.readline()
    if qpt_line == terminator:
        return None

    i_qpt = int(re.findall(_qpt_index_pattern, qpt_line)[0])
    floats = re.findall(_qpt_float_pattern, qpt_line)
    qpt = np.array([float(x) for x in floats[:3]])
    qweight = float(floats[3])

    direction: Optional[np.ndarray] = None
    if len(floats) >= 6:
        direction = floats[4:7]

    freq_lines = [f.readline().split()
                  for i in range(n_branches)]
    freq_col = 1
    qfreq = np.array([float(line[freq_col]) for line in freq_lines])
    extra: Optional[np.ndarray] = None
    if extra_columns is not None:
        if len(qpt_line.split()) > 6:
            # Indicates a split gamma point in .phonon, so there is an
            # additional column
            extra_columns = [x + 1 for x in extra_columns]
        extra = np.zeros((len(extra_columns), n_branches))
        for i, col in enumerate(extra_columns):
            extra[i] = np.array(
                [float(line[freq_col + col + 1]) for line in freq_lines])
    return _FrequencyBlock(i_qpt, qpt, direction, qweight, qfreq, extra)


def _read_frequency_blocks(*args, **kwargs) -> Iterator[_FrequencyBlock]:
    """Iterate over frequency blocks"""
    while True:
        frequency_block = _read_frequency_block(*args, **kwargs)
        if frequency_block is None:
            break
        yield frequency_block


def read_interpolation_data(
        filename: str,
        cell_vectors_unit: str = 'angstrom',
        atom_mass_unit: str = 'amu',
        force_constants_unit: str = 'hartree/bohr**2',
        born_unit: str = 'e',
        dielectric_unit: str = '(e**2)/(bohr*hartree)') -> Dict[str, Any]:
    """
    Reads data from a .castep_bin or .check file and returns it in a
    dictionary

    Parameters
    ----------
    filename
        The path and name of the file to read
    cell_vectors_unit
        The unit to return the cell vectors in
    atom_mass_unit
        The unit to return the atom masses in
    force_constants_unit
        The unit to return the force constants in
    born_unit
        The unit to return the Born charges in
    dielectric_unit
        The unit to return the dielectric permittivity tensor in

    Returns
    -------
    data_dict : dict
        A dict with the following keys: 'n_atoms', 'cell_vectors',
        'cell_vectors_unit', 'atom_r', 'atom_type', 'atom_mass',
        'atom_mass_unit', 'force_constants', 'force_constants_unit',
        'sc_matrix' and 'cell_origins'. Also contains 'born',
        'born_unit', 'dielectric' and 'dielectric_unit' if they are
        present in the .castep_bin or .check file.
    """
    with open(filename, 'rb') as f:
        int_type = '>i4'
        float_type = '>f8'
        first_cell_read = True
        while (header := _read_entry(f).strip()) != b'END':
            if header == b'BEGIN_UNIT_CELL':
                # CASTEP writes the cell twice: the first is the
                # geometry optimised cell, the second is the original
                # cell. We only want the geometry optimised cell.
                if first_cell_read:
                    (n_atoms, cell_vectors, atom_r, atom_mass,
                     atom_type) = _read_cell(f, int_type, float_type)
                    first_cell_read = False
            elif header == b'FORCE_CON':
                sc_matrix = np.transpose(np.reshape(
                    _read_entry(f, int_type), (3, 3)))
                n_cells_in_sc = int(np.rint(np.absolute(
                    np.linalg.det(sc_matrix))))
                # Transpose and reshape fc so it is indexed [nc, i, j]
                force_constants = np.ascontiguousarray(np.transpose(
                    np.reshape(_read_entry(f, float_type),
                               (n_cells_in_sc, 3*n_atoms, 3*n_atoms)),
                    axes=[0, 2, 1]))
                cell_origins = np.reshape(
                    _read_entry(f, int_type), (n_cells_in_sc, 3))
                _ = _read_entry(f, int_type)  # FC row not used
            elif header == b'BORN_CHGS':
                born = np.reshape(
                    _read_entry(f, float_type), (n_atoms, 3, 3))
            elif header == b'DIELECTRIC':
                dielectric = np.transpose(np.reshape(
                    _read_entry(f, float_type), (3, 3)))

    data_dict: Dict[str, Any] = {}
    cry_dict = data_dict['crystal'] = {}
    cry_dict['n_atoms'] = n_atoms
    cry_dict['cell_vectors'] = cell_vectors*ureg(
        'bohr').to(cell_vectors_unit).magnitude
    cry_dict['cell_vectors_unit'] = cell_vectors_unit
    # Normalise atom coordinates
    cry_dict['atom_r'] = atom_r - np.floor(atom_r)
    cry_dict['atom_type'] = atom_type
    cry_dict['atom_mass'] = atom_mass*ureg(
        'electron_mass').to(atom_mass_unit).magnitude
    cry_dict['atom_mass_unit'] = atom_mass_unit

    # Set entries relating to 'FORCE_CON' block
    try:
        data_dict['force_constants'] = force_constants*ureg(
            'hartree/bohr**2').to(force_constants_unit).magnitude
        data_dict['force_constants_unit'] = force_constants_unit
        data_dict['sc_matrix'] = sc_matrix
        data_dict['cell_origins'] = cell_origins
    except NameError:
        raise RuntimeError((
            f'Force constants matrix could not be found in {filename}. '
            f'\nEnsure PHONON_WRITE_FORCE_CONSTANTS: true and a '
            f'PHONON_FINE_METHOD has been chosen when running CASTEP'))

    # Set entries relating to dipoles
    try:
        data_dict['born'] = born*ureg('e').to(born_unit).magnitude
        data_dict['born_unit'] = born_unit
        data_dict['dielectric'] = dielectric*ureg(
            'e**2/(hartree*bohr)').to(dielectric_unit).magnitude
        data_dict['dielectric_unit'] = dielectric_unit
    except UnboundLocalError:
        pass

    return data_dict


def _read_cell(file_obj: BinaryIO, int_type: str, float_type: str
               ) -> Tuple[int, np.ndarray, np.ndarray,
                          np.ndarray, np.ndarray]:
    """
    Read cell data from a .castep_bin or .check file

    Parameters
    ----------
    file_obj
        File object in read mode for the .castep_bin or .check file
    int_type
        Python struct format string describing the size and endian-ness
        of ints in the file
    float_type
        Python struct format string describing the size and endian-ness
        of floats in the file

    Returns
    -------
    n_atoms
        Number of atoms in the unit cell
    cell_vectors
        Shape (3, 3) float ndarray. The unit cell vectors in bohr
    atom_r
        Shape (n_atoms, 3) float ndarray. The fractional position of
        each atom within the unit cell
    atom_mass
        Shape (n_atoms,) float ndarray. The mass of each atom in the
        unit cell in units of electron mass
    atom_type
        Shape (n_atoms,) string ndarray. The chemical symbols of each
        atom in the unit cell
    """
    while (header := _read_entry(file_obj).strip()) != b'END_UNIT_CELL':
        if header == b'CELL%NUM_IONS':
            n_atoms = _read_entry(file_obj, int_type)
        elif header == b'CELL%REAL_LATTICE':
            cell_vectors = np.transpose(np.reshape(
                _read_entry(file_obj, float_type), (3, 3)))
        elif header == b'CELL%NUM_SPECIES':
            n_species = _read_entry(file_obj, int_type)
        elif header == b'CELL%NUM_IONS_IN_SPECIES':
            n_atoms_in_species = _read_entry(file_obj, int_type)
            if n_species == 1:
                n_atoms_in_species = np.array([n_atoms_in_species])
        elif header == b'CELL%IONIC_POSITIONS':
            max_atoms_in_species = max(n_atoms_in_species)
            atom_r_tmp = np.reshape(_read_entry(file_obj, float_type),
                                    (n_species, max_atoms_in_species, 3))
        elif header == b'CELL%SPECIES_MASS':
            atom_mass_tmp = _read_entry(file_obj, float_type)
            if n_species == 1:
                atom_mass_tmp = np.array([atom_mass_tmp])
        elif header == b'CELL%SPECIES_SYMBOL':
            # Need to decode binary string for Python 3 compatibility
            if n_species == 1:
                atom_type_tmp = [_read_entry(file_obj, 'S8')
                                 .strip().decode('utf-8')]
            else:
                atom_type_tmp = [x.strip().decode('utf-8')
                                 for x in _read_entry(file_obj, 'S8')]
    # Get atom_r in correct form
    # CASTEP stores atom positions as 3D array (3,
    # max_atoms_in_species, n_species) so need to slice data to get
    # correct information
    atom_begin = np.insert(np.cumsum(n_atoms_in_species[:-1]), 0, 0)
    atom_end = np.cumsum(n_atoms_in_species)
    atom_r = np.zeros((n_atoms, 3))
    for i in range(n_species):
        atom_r[atom_begin[i]:atom_end[i], :] = atom_r_tmp[
            i, :n_atoms_in_species[i], :]
    # Get atom_type in correct form
    atom_type = np.array([])
    atom_mass = np.array([])
    for at in range(n_species):
        atom_type = np.append(
            atom_type,
            [atom_type_tmp[at] for i in range(n_atoms_in_species[at])])
        atom_mass = np.append(
            atom_mass,
            [atom_mass_tmp[at] for i in range(n_atoms_in_species[at])])

    return n_atoms, cell_vectors, atom_r, atom_mass, atom_type


def _read_entry(file_obj: BinaryIO, dtype: str = ''
                ) -> Union[str, int, float, np.ndarray]:
    """
    Read a record from a Fortran binary file, including the beginning
    and end record markers and return the data inbetween

    Parameters
    ----------
    f
        File object in read mode for the Fortran binary file
    dtype
        String determining what order and type to unpack the bytes as.
        See 'Format Strings' in Python struct documentation

    Returns
    -------
    data
        Data type returned depends on dtype specified. If dtype is not
        specified, return type is a string. If there is more than one
        element in the record, it is returned as an ndarray of floats or
        integers

    """
    def record_mark_read(file_obj):
        # Read 4 byte Fortran record marker
        rawdata = file_obj.read(4)
        if rawdata == b'':
            raise EOFError(
                'Problem reading binary file: unexpected EOF reached')
        return struct.unpack('>i', rawdata)[0]

    begin = record_mark_read(file_obj)
    if dtype:
        n_bytes = int(dtype[-1])
        n_elems = int(begin/n_bytes)
        if n_elems > 1:
            data = np.fromfile(file_obj, dtype=dtype, count=n_elems)
            if 'i' in dtype:
                data = data.astype(np.int32)
            elif 'f' in dtype:
                data = data.astype(np.float64)
        else:
            if 'i' in dtype:
                data = struct.unpack('>i', file_obj.read(begin))[0]
            elif 'f' in dtype:
                data = struct.unpack('>d', file_obj.read(begin))[0]
            else:
                data = file_obj.read(begin)
    else:
        data = file_obj.read(begin)
    end = record_mark_read(file_obj)
    if begin != end:
        raise IOError("""Problem reading binary file: beginning and end
                         record markers do not match""")

    return data



# -*- coding: UTF-8 -*-
from argparse import ArgumentParser, Namespace, SUPPRESS
import itertools
from pathlib import Path
from typing import List, Tuple, Optional, Sequence

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from matplotlib.lines import Line2D

from euphonic import ForceConstants, Spectrum1D, ureg
from euphonic.brille import BrilleInterpolator
from euphonic.cli.utils import (_get_cli_parser, get_args,
                                _brille_calc_modes_kwargs,
                                load_data_from_file,
                                _get_energy_bins)
from euphonic.sampling import recurrence_sequence


def main(params: Optional[List[str]] = None) -> None:
    args = get_args(get_parser(), params)
    params = vars(args)
    check_brille_settings(**params)


def check_brille_settings(
        filename: str,
        npts: int = 500,
        use_brille: bool = True,
        brille_grid_type: str = 'trellis',
        brille_npts: int = 5000,
        brille_npts_density: Optional[int] = None,
        n: int = 0,
        energy_broadening: Optional[float] = None,
        ebins: int = 500,
        e_min: Optional[float] = None,
        e_max: Optional[float] = None,
        energy_unit: str = 'meV',
        shape: str = 'gauss',
        **calc_modes_kwargs
        ) -> None:

    fc = load_data_from_file(filename)
    if not isinstance(fc, ForceConstants):
        raise TypeError('Force constants are required to use the '
                        'euphonic-check-brille-settings tool')

    brille_calc_modes_kwargs = _brille_calc_modes_kwargs(
        Namespace(**calc_modes_kwargs))
    qpts = np.fromiter(itertools.chain.from_iterable(
        recurrence_sequence(npts, order=3)), dtype=float).reshape(-1, 3)

    # Calculate Euphonic frequencies, structure factors and intensities
    modes = fc.calculate_qpoint_phonon_modes(qpts, **calc_modes_kwargs)
    modes.frequencies_unit = energy_unit
    sf = modes.calculate_structure_factor()
    ebins = _get_energy_bins(modes, ebins + 1, emin=e_min, emax=e_max)
    sqw_avg = sf.calculate_1d_average(ebins)
    if energy_broadening is not None:
        sqw_avg = sqw_avg.broaden(energy_broadening*ebins.units, shape=shape)

    # Set up axis labels
    frequency_label = f'Frequency / {modes.frequencies.units:~P}'
    residual_label = 'Epsilon / '
    fname = f'{Path(filename).name}'

    # Set up plot for frequency residuals
    _, ax_freq = plt.subplots()
    ax_freq.set_xlabel(frequency_label)
    ax_freq.set_ylabel(residual_label + f'{modes.frequencies.units:~P}')
    ax_freq.set_title(f'Frequency residual ({fname})')

    # Set up plot for structure factor residuals
    _, ax_sf = plt.subplots(subplot_kw={'projection': '3d'})
    ax_sf.set_xlabel(frequency_label)
    ax_sf.set_ylabel(f'Structure factors / {sf.structure_factors.units:~P}')
    ax_sf.set_zlabel(residual_label + f'{sf.structure_factors.units:~P}')
    ax_sf.set_title(f'Structure factor residual ({fname})')

    # Set up plot for 1D average intensity & residuals
    ax_avg, eu_handle = create_intensity_plot(
        sqw_avg, f'1D Average Intensity over all Q-points ({fname})',
        frequency_label, residual_label)

    # Set up plot for q-point specific intensity & residuals
    ax_sqwn = []
    sqw1d_n = []
    if n > 0:
        sqw = sf.calculate_sqw_map(ebins)
        if energy_broadening is not None:
            sqw = sqw.broaden(y_width=energy_broadening*ebins.units,
                              shape=shape)
        for idx in range(n):
            # Use idx + 1 to avoid Q=[0., 0., 0.] at idx=0
            sqw1d = Spectrum1D(sqw.y_data, sqw.z_data[idx + 1])
            sqw1d_n.append(sqw1d)
            title = f'Intensity at Q={np.round(qpts[idx + 1], 4)} ({fname})'
            ax, _ = create_intensity_plot(sqw1d, title, frequency_label,
                                          residual_label)
            ax_sqwn.append(ax)

    if brille_npts_density is not None:
        brille_params = brille_npts_density
        brille_param_key = 'grid_density'
    else:
        brille_params = brille_npts
        brille_param_key = 'grid_npts'

    handles = []
    for brille_param_i in brille_params:

        # Calculate Brille frequencies, structure factors and intensities
        brille_fc = BrilleInterpolator.from_force_constants(
            fc, grid_type=brille_grid_type,
            interpolation_kwargs=calc_modes_kwargs,
            **{brille_param_key: brille_param_i})
        brille_modes = brille_fc.calculate_qpoint_phonon_modes(
            qpts, **brille_calc_modes_kwargs)
        brille_modes.frequencies_unit = energy_unit
        brille_sf = brille_modes.calculate_structure_factor()
        brille_sqw_avg = brille_sf.calculate_1d_average(ebins)
        if energy_broadening is not None:
            brille_sqw_avg = brille_sqw_avg.broaden(
                energy_broadening*ebins.units, shape=shape)

        # Create plot labels
        actual_brille_npts = len(brille_fc._grid.rlu)
        bz_vol = brille_fc._grid.BrillouinZone.ir_polyhedron.volume
        actual_brille_density = int(actual_brille_npts/bz_vol)
        label = f'{actual_brille_npts} ({actual_brille_density})'

        # Plot frequency residuals
        eps = (brille_modes.frequencies
               - modes.frequencies).magnitude.flatten()
        ax_freq.plot(modes.frequencies.magnitude.flatten(), eps, 'x',
                     label=label)

        # Plot structure factor residuals
        eps_sf = (brille_sf.structure_factors
                  - sf.structure_factors).magnitude.flatten()
        ax_sf.plot(modes.frequencies.magnitude.flatten(),
                   sf.structure_factors.magnitude.flatten(), eps_sf, 'x',
                   label=label)

        # Plot 1D average intensity & residuals
        handle = plot_to_intensity_axes(ax_avg, sqw_avg, brille_sqw_avg, label)
        handles.append(handle)

        # Plot q-point specific intensity & residuals
        if n > 0:
            brille_sqw = brille_sf.calculate_sqw_map(ebins)
            if energy_broadening is not None:
                brille_sqw = brille_sqw.broaden(
                    y_width=energy_broadening*ebins.units, shape=shape)
            for idx, ax_sqw in enumerate(ax_sqwn):
                brille_sqw1d = Spectrum1D(brille_sqw.y_data,
                                          brille_sqw.z_data[idx + 1])
                plot_to_intensity_axes(
                    ax_sqw, sqw1d_n[idx], brille_sqw1d, label)

    # Add legends
    legend_title = (f'Brille mesh: N qpts '
                    f'(density {ureg("angstrom**3").units:~P})')
    ax_freq.legend(title=legend_title)
    ax_sf.legend(title=legend_title)
    add_intensity_legend(ax_avg, eu_handle, handles, legend_title)
    for ax_sqw in ax_sqwn:
        add_intensity_legend(ax_sqw, eu_handle, handles, legend_title)

    plt.show()


def create_intensity_plot(sqw1d: Spectrum1D, title: str,
                          frequency_label: str, residual_label: str,
                          ) -> Tuple[Tuple[Axes, Axes], Line2D]:
    fig, axes = plt.subplots(nrows=2, sharex=True)
    axes[1].set_xlabel(frequency_label)
    axes[1].set_ylabel(residual_label + f'{sqw1d.y_data.units:~P}')
    axes[0].set_ylabel(f'Intensity / {sqw1d.y_data.units:~P}')
    ebins_plot = sqw1d.get_bin_centres().magnitude
    handle, = axes[0].plot(ebins_plot, sqw1d.y_data.magnitude,
                           label='Euphonic', color='k')
    fig.suptitle(title)
    fig.tight_layout()
    return axes, handle


def plot_to_intensity_axes(axes: Tuple[Axes, Axes], sqw1d: Spectrum1D,
                           brille_sqw1d: Spectrum1D, label: str) -> Line2D:
    eps = (brille_sqw1d.y_data - sqw1d.y_data).magnitude
    ebins_plot = sqw1d.get_bin_centres().magnitude
    handle, = axes[0].plot(ebins_plot, brille_sqw1d.y_data.magnitude,
                           label=label)
    axes[1].plot(ebins_plot, eps)
    return handle


def add_intensity_legend(axes: Tuple[Axes, Axes], euphonic_handle: Line2D,
                         brille_handles: Sequence[Line2D], title: str
                         ) -> None:
    euph_legend = axes[0].legend(handles=[euphonic_handle], loc='upper left')
    axes[0].add_artist(euph_legend)
    axes[0].legend(handles=brille_handles, loc='upper right', title=title)


def get_parser() -> ArgumentParser:
    parser, sections = _get_cli_parser(
        features={'read-fc', 'brille', 'ebins'},
        conflict_handler='resolve')
    parser.description=(
        'Estimate linear interpolation error by comparing with Fourier '
        'interpolation')

    sections['brille'].add_argument('--npts', type=int, default=1000,
                                    help=('Number of quasirandom samples '
                                          'used in error estimate.'))
    sections['brille'].add_argument('--use-brille', help=SUPPRESS)
    sections['brille'].add_argument(
        '--brille-npts', type=int, default=[5000], nargs='+',
        help=('Approximate number of q-points to generate on the '
              'Brille grid, is passed to the "n_grid_points" kwarg '
              'of BrilleInterpolator.from_force_constants'))
    sections['brille'].add_argument(
        '--brille-npts-density', type=int, default=None, nargs='+',
        help=('Approximate density of q-points to generate on the '
              'Brille grid, is passed to the "grid_density" kwarg '
              'of BrilleInterpolator.from_force_constants'))
    sections['brille'].add_argument(
        '-n', type=int, default=0,
        help=('Number of single q-point intensity slices to plot'))

    return parser



"""Matplotlib stylesheets for plot styling"""
from importlib_resources import files

base_style = files(__package__) / "base.mplstyle"
intensity_widget_style = files(__package__) / "intensity_widget.mplstyle"



from argparse import ArgumentParser
from math import ceil
from typing import Callable, Dict, List, Optional, Sequence, Union

import matplotlib.style
import numpy as np
from numpy.polynomial import Polynomial
from pint import Unit


from euphonic import ureg, ForceConstants, Quantity
from euphonic.cli.utils import (_calc_modes_kwargs, _brille_calc_modes_kwargs,
                                _compose_style, _get_cli_parser,
                                _get_debye_waller, _get_energy_bins,
                                _get_q_distance, _get_pdos_weighting,
                                _arrange_pdos_groups, _plot_label_kwargs)
from euphonic.cli.utils import (load_data_from_file, get_args,
                                matplotlib_save_or_show)
import euphonic.plot
from euphonic.powder import (sample_sphere_dos, sample_sphere_pdos,
                             sample_sphere_structure_factor)
from euphonic.spectra import apply_kinematic_constraints
from euphonic.styles import base_style, intensity_widget_style
import euphonic.util

# Dummy tqdm function if tqdm progress bars unavailable
try:
    from tqdm.auto import tqdm
except ModuleNotFoundError:
    try:
        from tqdm import tqdm
    except ModuleNotFoundError:
        def tqdm(sequence):
            return sequence


def get_parser() -> ArgumentParser:

    parser, sections = _get_cli_parser(
        features={'read-fc', 'pdos-weighting', 'ins-weighting',
                  'powder', 'plotting', 'ebins', 'q-e', 'map',
                  'brille', 'kinematic', 'scaling'})

    sections['q'].description = (
        '"GRID" options relate to Monkhorst-Pack sampling for the '
        'Debye-Waller factor, and only apply when --weighting=coherent and '
        '--temperature is set. "NPTS" options determine spherical groups of '
        'q-points for powder-averaging. "Q" options relate to the sphere '
        'sizes (i.e. radial distances).')

    sections['q'].add_argument('--q-min', type=float, default=0., dest='q_min',
                               help="Minimum |q| in 1/LENGTH_UNIT")
    sections['q'].add_argument('--q-max', type=float, default=3., dest='q_max',
                               help="Maximum |q| in 1/LENGTH_UNIT")
    sections['plotting'].add_argument('--no-widgets', action='store_true',
                                      dest='disable_widgets', default=False,
                                      help=("Don't use Matplotlib widgets to "
                                            "enable interactive setting of "
                                            "colormap intensity limits"))

    kinematic = parser.add_argument_group('Kinematic constraints')
    e_lims = kinematic.add_mutually_exclusive_group()

    e_lims.add_argument('--e-incident', '--e-i', dest='e_i',
                        type=float, default=None,
                        help="Incident energy for direct-geometry constraints")
    e_lims.add_argument('--e-final', '--e-f', dest='e_f',
                        type=float, default=None,
                        help="Final energy for indirect-geometry constraints")
    kinematic.add_argument(
        '--angle-range', nargs=2, type=float, dest='angle_range',
        default=[0., 180.],
        help=("Range of scattering angles (2θ) in degrees. These lower/upper "
              "bounds are used with incident/final energy to determine "
              "accessible (|q|, ω) region."))
    return parser


def _get_broaden_kwargs(q_broadening: Optional[Sequence[float]] = None,
                        q_unit: Unit = ureg('1/angstrom').units,
                        energy_broadening: Optional[Sequence[float]] = None,
                        energy_unit: Unit = ureg('meV').units,
                        ) -> Dict[str, Union[Quantity, Callable, None]]:
    """Collect suitable width arguments for 2D broaden() method

    Internal Euphonic methods take a Quantity or a callable returning Quantity:
    CLI takes numbers and units separately so these have to be collected into
    appropriate objects.
    """

    if (q_broadening is not None) and len(q_broadening) > 1:
        q_poly = Polynomial(q_broadening)

        def q_width(x):
            return q_poly(x.to(q_unit).magnitude
                          ) * q_unit
    elif q_broadening:
        q_width = q_broadening[0] * q_unit
    else:
        q_width = None

    if (energy_broadening is not None) and len(energy_broadening) > 1:
        energy_poly = Polynomial(energy_broadening)

        def energy_width(x):
            return energy_poly(x.to(energy_unit).magnitude
                               ) * ureg(energy_unit)
    elif energy_broadening:
        energy_width = energy_broadening[0] * ureg(energy_unit)
    else:
        energy_width = None

    return dict(x_width=q_width, y_width=energy_width)


def main(params: Optional[List[str]] = None) -> None:
    args = get_args(get_parser(), params)
    calc_modes_kwargs = _calc_modes_kwargs(args)

    # Make sure we get an error if accessing NPTS inappropriately
    if args.npts_density is not None:
        args.npts = None

    fc = load_data_from_file(args.filename, verbose=True)
    if not isinstance(fc, ForceConstants):
        raise TypeError('Force constants are required to use the '
                        'euphonic-powder-map tool')
    if args.pdos is not None and args.weighting == 'coherent':
        raise ValueError('"--pdos" is only compatible with '
                         '"--weighting" options that include dos')

    if args.use_brille:
        from euphonic.brille import BrilleInterpolator
        fc = BrilleInterpolator.from_force_constants(
            fc, grid_type=args.brille_grid_type,
            grid_npts=args.brille_npts,
            grid_density=args.brille_npts_density,
            interpolation_kwargs=calc_modes_kwargs)
        calc_modes_kwargs = _brille_calc_modes_kwargs(args)

    print("Setting up dimensions...")
    q_min = _get_q_distance(args.length_unit, args.q_min)
    q_max = _get_q_distance(args.length_unit, args.q_max)
    recip_length_unit = q_min.units

    n_q_bins = ceil((args.q_max - args.q_min) / args.q_spacing)
    q_bin_edges = np.linspace(q_min.magnitude, q_max.magnitude, n_q_bins + 1,
                              endpoint=True) * recip_length_unit
    q_bin_centers = (q_bin_edges[:-1] + q_bin_edges[1:]) / 2

    # Use X-point modes to estimate frequency range, set up energy bins
    # (Not Gamma in case there are only 3 branches; value would be zero!)
    modes = fc.calculate_qpoint_frequencies(
        np.array([[0., 0., 0.5]]), **calc_modes_kwargs)
    modes.frequencies_unit = args.energy_unit

    if args.e_i is not None and args.e_max is None:
        emax = args.e_i
    else:
        emax = args.e_max

    energy_bins = _get_energy_bins(
        modes, args.ebins + 1, emin=args.e_min, emax=emax,
        headroom=1.2)  # Generous headroom as we only checked one q-point

    if args.weighting in ('coherent',):
        # Compute Debye-Waller factor once for re-use at each mod(q)
        # (If temperature is not set, this will be None.)
        if args.temperature is not None:
            temperature = args.temperature * ureg('K')
            dw = _get_debye_waller(temperature, fc, grid=args.grid,
                                   grid_spacing=(args.grid_spacing
                                                 * recip_length_unit),
                                   **calc_modes_kwargs)
        else:
            temperature = None
            dw = None

    print(f"Sampling {n_q_bins} |q| shells between {q_min:~P} and {q_max:~P}")

    z_data = np.empty((n_q_bins, len(energy_bins) - 1))

    for q_index in tqdm(range(n_q_bins)):
        q = q_bin_centers[q_index]

        if args.npts_density is not None:
            npts = ceil(args.npts_density * (q / recip_length_unit)**2)
            npts = max(args.npts_min,
                       min(args.npts_max, npts))
        else:
            npts = args.npts

        if args.weighting == 'dos' and args.pdos is None:
            spectrum_1d = sample_sphere_dos(
                fc, q,
                npts=npts, sampling=args.sampling, jitter=args.jitter,
                energy_bins=energy_bins,
                **calc_modes_kwargs)
        elif 'dos' in args.weighting:
            spectrum_1d_col = sample_sphere_pdos(
                    fc, q,
                    npts=npts, sampling=args.sampling, jitter=args.jitter,
                    energy_bins=energy_bins,
                    weighting=_get_pdos_weighting(args.weighting),
                    **calc_modes_kwargs)
            spectrum_1d = _arrange_pdos_groups(spectrum_1d_col, args.pdos)
        elif args.weighting == 'coherent':
            spectrum_1d = sample_sphere_structure_factor(
                fc, q,
                dw=dw,
                temperature=temperature,
                sampling=args.sampling, jitter=args.jitter,
                npts=npts,
                energy_bins=energy_bins,
                **calc_modes_kwargs)

        z_data[q_index, :] = spectrum_1d.y_data.magnitude

    print(f"Final npts: {npts}")

    spectrum = euphonic.Spectrum2D(q_bin_edges, energy_bins,
                                   z_data * spectrum_1d.y_data.units)

    if (args.energy_broadening is not None) or (args.q_broadening is not None):
        spectrum = spectrum.broaden(
            shape=args.shape,
            **_get_broaden_kwargs(q_broadening=args.q_broadening,
                                  q_unit=recip_length_unit,
                                  energy_broadening=args.energy_broadening,
                                  energy_unit=args.energy_unit))

    if not (args.e_i is None and args.e_f is None):
        print("Applying kinematic constraints")
        energy_unit = args.energy_unit
        e_i = args.e_i * ureg(energy_unit) if (args.e_i is not None) else None
        e_f = args.e_f * ureg(energy_unit) if (args.e_f is not None) else None
        spectrum = apply_kinematic_constraints(
            spectrum, e_i=e_i, e_f=e_f, angle_range=args.angle_range)

    if args.scale is not None:
        spectrum *= args.scale

    print(f"Plotting figure: max intensity "
          f"{np.nanmax(spectrum.z_data.magnitude) * spectrum.z_data.units:~P}")
    plot_label_kwargs = _plot_label_kwargs(
        args, default_xlabel=f"|q| / {q_min.units:~P}",
        default_ylabel=f"Energy / {spectrum.y_data.units:~P}")

    if args.save_json:
        spectrum.to_json_file(args.save_json)
    if args.disable_widgets:
        base = [base_style]
    else:
        base = [base_style, intensity_widget_style]
    style = _compose_style(user_args=args, base=base)
    with matplotlib.style.context(style):
        fig = euphonic.plot.plot_2d(spectrum,
                                    vmin=args.vmin,
                                    vmax=args.vmax,
                                    **plot_label_kwargs)

        if args.disable_widgets is False:
            # TextBox only available from mpl 2.1.0
            try:
                from matplotlib.widgets import TextBox
            except ImportError:
                args.disable_widgets = True

        if args.disable_widgets is False:
            min_label = f'Min Intensity ({spectrum.z_data.units:~P})'
            max_label = f'Max Intensity ({spectrum.z_data.units:~P})'
            boxw = 0.15
            boxh = 0.05
            x0 = 0.1 + len(min_label)*0.01
            y0 = 0.025
            axmin = fig.add_axes([x0, y0, boxw, boxh])
            axmax = fig.add_axes([x0, y0 + 0.075, boxw, boxh])
            image = fig.get_axes()[0].images[0]
            cmin, cmax = image.get_clim()
            pad = 0.05
            fmt_str = '.2e' if cmax < 0.1 else '.2f'
            minbox = TextBox(axmin, min_label,
                             initial=f'{cmin:{fmt_str}}', label_pad=pad)
            maxbox = TextBox(axmax, max_label,
                             initial=f'{cmax:{fmt_str}}', label_pad=pad)

            def update_min(min_val):
                image.set_clim(vmin=float(min_val))
                fig.canvas.draw()

            def update_max(max_val):
                image.set_clim(vmax=float(max_val))
                fig.canvas.draw()
            minbox.on_submit(update_min)
            maxbox.on_submit(update_max)

        matplotlib_save_or_show(save_filename=args.save_to)
